#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dll.h"
#include "madcap.h"
#include "discover.h"

#include <time.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <readline/readline.h>
#include <readline/history.h>

/* Start of default values, change if needed */
static uint16 min_desired_addresses = 1;
static uint16 max_desired_addresses = 1;

/* 0 = 'now' */
static int min_desired_start_hour = 0;
static int max_desired_start_hour = 0;
static int min_desired_start_minute = 0;
static int max_desired_start_minute = 0;

static int start_year = 101;/*2001*/
static int start_second = 0;
static int month_day = 0;
static int month = 0;
/* end default values */

/* values passed to the API functions */   
static uint32 min_desired_start_time;
static uint32 max_desired_start_time;
static uint32 min_desired_lifetime = 0;    /* seconds */
static uint32 max_desired_lifetime = 1800; /* seconds */

/* Functions, user-available and directly mapped to the API.
 * These function names are the same as the command names that can be
 * used at the command line to execute the API call.
 * General function for allocationg addresses and server known
 */
static int alloc_server();

/* General function for allocationg addresses and server not known */
static int alloc_noserver();

/* Allocating only a single address */
static int alloc_one();

/* Allocate addresses with a single, fixed lifetime */
static int alloc_exact();

/* Allocate addresses for immediate use */
static int alloc_imm();

/* Changing addresses' lifetime */
static int change_lifetime();

/* Deallocating addresses */
static int deallocate();

/* View/set the scope */
static int list_scope(void);
static int set_scope(void);

/* Functions used to view/modify the values passed to the API functions. */
/* Functions to view/set server to locate */
static int set_server();
static int list_server();
static int clear_server();

/*functions for view/set min/max number of wanted addresses */
static void list_desired_addresses();
static int set_desired_addresses();

/* functions for view/set times for start/lifetime */
static void list_desired_times();
static int set_desired_times();

/* user "help" */
static void help();

/* lists all allocation-sessions */
static int list_leases(void);
/* sets active allocation-session */
static int set_lease(void);

/* Internal */
/* functions for the time values */
static int set_default_values();

/* Set default server(localhost) */
static int set_default_server();

/* functions for default number of addresses */
static uint16 get_min_desired_addresses();
static uint16 get_max_desired_addresses();

/* set the default IPv4 local scope */
static int set_default_scope(void);

static uint16 IPV4 = 1;    
static char last_cmd[20];
static struct hostent *host_entry;
static struct in_addr *addr_list;
static int validserver = 0;   /* keep track if server is entered */
static uint32 server_address; /* current server address */
static int allocID = 1;       /* counts the number of different 'jobs' */

typedef struct allocation{
    int idofallocation;
    struct lease *le; 
}allocation;

/* holds the current selected lease(when an alloction has been made) */
static allocation *currentLease;

/* holds all leases */
static DLL *dll;

/* Represents a multicast scope. */
static struct scope sc;

/* Represents a name, for use in the scopes. */
static struct name names;

/* An address range. The range is [start, end]. */
static struct address_range ar;

#ifndef MC_FreeBSD
extern int gethostname(char *name, size_t len);
#endif

static char *cmd = (char *)NULL;

int main(int argc, char *argv[]){
      
    set_default_values();
    set_default_scope();
    set_default_server();
    dll = createDLL();   
    
    while(1){
        cmd = readline("Enter cmd: ");
        add_history(cmd);
       
        if(strcmp(cmd,"1")==0){ 
            strcpy(cmd,last_cmd);
        }else{           
            strcpy(last_cmd,cmd); /* save last command */
        }

        if (strcmp(cmd,"alloc_server")==0){
            alloc_server();
        }else if(strcmp(cmd,"alloc_noserver")==0){   
            alloc_noserver();
        }else if(strcmp(cmd,"alloc_one")==0){   
            alloc_one();
        }else if(strcmp(cmd,"alloc_exact")==0){      
            alloc_exact();
        }else if(strcmp(cmd,"alloc_imm")==0){      
            alloc_imm();

        }else if(strcmp(cmd,"change_lifetime")==0){   
            change_lifetime();    
        }else if(strcmp(cmd,"deallocate")==0){   
            deallocate();
                        
        }else if(strcmp(cmd,"list_scope")==0){   
            list_scope();
        }else if(strcmp(cmd,"set_scope")==0){   
            set_scope(); 
            
        }else if(strcmp(cmd,"list_server")==0){   
            list_server();
        }else if(strcmp(cmd,"set_server")==0){   
            set_server();
        }else if(strcmp(cmd,"clear_server")==0){   
            clear_server();            
            
        }else if(strcmp(cmd,"set_addresses")==0){
            set_desired_addresses();
        }else if(strcmp(cmd,"list_addresses")==0){
            list_desired_addresses();            
            
        }else if(strcmp(cmd,"set_times")==0){
            set_desired_times();
        }else if(strcmp(cmd,"list_times")==0){
            list_desired_times();            
        }else if(strcmp(cmd,"show")==0){
            list_desired_addresses();
            list_desired_times();            
            
        }else if(strcmp(cmd,"list_leases")==0){
            list_leases();
        }else if(strcmp(cmd,"set_lease")==0){
            set_lease();

        }else if(strcmp(cmd,"discover")==0){
            discover(sc);            
            
        }else if(strcmp(cmd,"help")==0){
            help(0);
        }else if(strcmp(cmd,"exit")==0){
            exit(0);        
        }else{
            printf("Unknown cmd: %s\n",cmd);
        }
    }
    return 0;
}

/* Can be used when the server is known. */
static int alloc_server(){
    allocation *tmp;
    int ret_val  = 0;
    int status   = 0;
    
    tmp                = malloc(sizeof(struct allocation));
    tmp->le            = malloc(sizeof(lease));
    tmp->le->addresses = NULL; 
    tmp->le->lease_id  = NULL; 
    
    
    if (!validserver){
        /* set server */
        if (set_server()==-1)
            return -1;
    }
    printf("Connecting to ");
    list_server();
    ret_val = alloc_multicast_addr(IPV4,
                                   &sc,
                                   min_desired_addresses,
                                   max_desired_addresses,
                                   min_desired_start_time,
                                   max_desired_start_time,
                                   min_desired_lifetime,
                                   max_desired_lifetime,
                                   server_address,
                                   tmp->le,
                                   &status);
       
    if (ret_val == -1){
        mc_perror("Could not allocate address(es)");
        return -1;  
    }else{
        tmp->idofallocation = allocID;
        allocID++;
        insertElement(dll,tmp,0);
        currentLease = tmp;       
    }
    return 0;
}


/* General function for allocationg addresses and server not known */
static int alloc_noserver(){ 
    allocation *tmp;
    int ret_val = 0;
    int status = 0;
    tmp                = malloc(sizeof(struct allocation));
    tmp->le            = malloc(sizeof(lease));
    tmp->le->addresses = malloc(sizeof(uint32));    
    tmp->le->lease_id  = malloc(sizeof(uint8));
        
    printf("Searching for server to connect to... \n");
    
    ret_val = alloc_noserv_addr(IPV4,
                                &sc,
                                min_desired_addresses,
                                max_desired_addresses,
                                min_desired_start_time,
                                max_desired_start_time,
                                min_desired_lifetime,
                                max_desired_lifetime,                                
                                tmp->le,
                                &status);    
    
    if (ret_val == -1){
        mc_perror("Could not allocate address(es)");
        return -1;
    }else{        
        tmp->idofallocation = allocID;
        allocID++;
        insertElement(dll,tmp,0);
        currentLease = tmp;        
    }
    return 0;
}
/* Allocating only a single address */
static int alloc_one(){
    allocation *tmp;
    int ret_val = 0;
    int status = 0;
    tmp                = malloc(sizeof(struct allocation));
    tmp->le            = malloc(sizeof(lease));
    tmp->le->addresses = malloc(sizeof(uint32));    
    tmp->le->lease_id  = malloc(sizeof(uint8));
       
    printf("Searching for server to connect to... \n");
    
    ret_val = alloc_one_addr(IPV4,
                             &sc,
                             min_desired_start_time,
                             max_desired_start_time,
                             min_desired_lifetime,
                             max_desired_lifetime,                                
                             tmp->le,
                             &status);    
    
    if (ret_val == -1){
        mc_perror("Could not allocate address(es)");
        return -1;
    }else{        
        tmp->idofallocation = allocID;
        allocID++;
        insertElement(dll,tmp,0);
        currentLease = tmp;        
    }
    return 0;
}


/* Allocate addresses with a single, fixed lifetime */
static int alloc_exact(){
    allocation *tmp;
    int ret_val = 0;
    int status = 0;
    tmp                = malloc(sizeof(struct allocation));
    tmp->le            = malloc(sizeof(lease));
    tmp->le->addresses = malloc(sizeof(uint32));    
    tmp->le->lease_id  = malloc(sizeof(uint8));
        
    printf("Searching for server to connect to... \n");
    
    ret_val = alloc_exact_addr(IPV4,
                               &sc,
                               min_desired_addresses,
                               max_desired_addresses,
                               min_desired_start_time,
                               max_desired_start_time,
                               min_desired_lifetime,
                               tmp->le,
                               &status);    
    
    if (ret_val == -1){
        mc_perror("Could not allocate address(es)");
        return -1;
    }else{         
        tmp->idofallocation = allocID;
        allocID++;
        insertElement(dll,tmp,0);
        currentLease        = tmp;        
    }
    return 0;
}

/* Allocate addresses for immediate use */
static int alloc_imm(){
    allocation *tmp;
    int ret_val = 0;
    int status = 0;
    tmp                = malloc(sizeof(struct allocation));
    tmp->le            = malloc(sizeof(lease));
    tmp->le->addresses = malloc(sizeof(uint32));    
    tmp->le->lease_id  = malloc(sizeof(uint8));
        
    printf("Searching for server to connect to... \n");
    
    ret_val = alloc_immediate_addr(IPV4,
                                   &sc,
                                   min_desired_addresses,
                                   max_desired_addresses,
                                   min_desired_lifetime,
                                   max_desired_lifetime,
                                   tmp->le,
                                   &status);                                
    
    if (ret_val == -1){       
        mc_perror("Could not allocate address(es)");
        return -1;
    }else{       
        tmp->idofallocation = allocID;
        allocID++;
        insertElement(dll,tmp,0);
        currentLease        = tmp;        
    }
    return 0;
}

/* Changing addresses' lifetime */
static int change_lifetime(){
    int ret_val = 0;
    if (currentLease == NULL){
        printf("No addresses allocated.\n");
        return -1;
    }
    ret_val = change_multicast_addr_lifetime((currentLease->le),
                                             min_desired_lifetime,
                                             max_desired_lifetime,
                                             &(currentLease->le->lifetime));
    if (ret_val == -1){
        mc_perror("Could not change multicast address(es) lifetime");
    }
    return ret_val;
}


/* set the server to locate when allocating addresses */
static int set_server(){
    struct hostent *host_entry_tmp;
    cmd = readline("Enter server name: ");
    add_history(cmd);
    
    host_entry_tmp = gethostbyname(cmd); 
    if (host_entry_tmp == NULL){
        fprintf(stderr,"Can not find host %s\n", cmd);
        return -1;
    }
    host_entry     = gethostbyname(cmd);
    addr_list      = (struct in_addr *) *host_entry->h_addr_list;
    server_address = inet_addr(inet_ntoa(*addr_list));
    /* convert to host byte order since inet_addr gives network byte order */
    server_address = ntohl(server_address);
    validserver    = 1;   
    return 0;
}


/* clear current server */
static int clear_server(){
    validserver = 0; 
    return 0;
}


/* show server */
static int list_server(){
    if (!validserver){
        printf("Server not set.\n");
    }else{
        printf("Server:  %s (%s)\n",
               host_entry->h_name,inet_ntoa(*addr_list));
    }
    return 0;
}


/* get min desired addresses */
static uint16 get_min_desired_addresses(){
    return min_desired_addresses;
}


/* get ax desired addresses */
static uint16 get_max_desired_addresses(){
    return max_desired_addresses;
}


/* sets default time values */
static int set_default_values(){
    struct tm tp;
    struct tm now;
    time_t t=0;

    /* get current time */
    time(&t);
    now = *localtime(&t);

    /* setting min_desired_start_time */
    tp.tm_hour = now.tm_hour + min_desired_start_hour;    
    tp.tm_min  = now.tm_min + min_desired_start_minute;
    tp.tm_sec  = now.tm_sec + start_second;
    tp.tm_mday = now.tm_mday + month_day;
    tp.tm_mon  = now.tm_mon  + month;
    tp.tm_year = now.tm_year;
    min_desired_start_time  = mktime(&tp);

    /* setting max_desired_start_time */
    tp.tm_hour = now.tm_hour + max_desired_start_hour;
    tp.tm_min  = now.tm_min + max_desired_start_minute;
    tp.tm_sec  = now.tm_sec + start_second;
    tp.tm_mday = now.tm_mday + month_day;
    tp.tm_mon  = now.tm_mon  + month;
    tp.tm_year = now.tm_year;
    max_desired_start_time  = mktime(&tp);   
    
    return 0;
}


/* Set default server(localhost) */
static int set_default_server(){
    char buf[100];
    int len; 
    len = 100;

    gethostname(buf, len);    
    host_entry     = gethostbyname(buf);
    addr_list      = (struct in_addr *) *host_entry->h_addr_list;
    server_address = inet_addr(inet_ntoa(*addr_list));
    /* convert to host byte order since inet_addr gives network byte order */
    server_address = ntohl(server_address);
    validserver    = 1;
    return 0;
}


/* set min & max number of desired addresses */
static int set_desired_addresses(){   
    cmd = readline("Set min desired addresses: ");
    add_history(cmd);
    min_desired_addresses = atoi(cmd);

    cmd = readline("Set max desired addresses: ");
    add_history(cmd);
    max_desired_addresses = atoi(cmd);     
    
    return 0;
}


/* print out desired addresses */
static void list_desired_addresses(){
    printf("Min desired addresses: \t%d\n",get_min_desired_addresses());
    printf("Max desired addresses: \t%d\n",get_max_desired_addresses());
}


/* print out desired times */
static void list_desired_times(){
    if(min_desired_start_time == 0) {
        printf("Min desired start time: %ld\n", min_desired_start_time);
    } else {
        printf("Min desired start time: %s",ctime(&min_desired_start_time));
    }
    if(max_desired_start_time == 0) {
        printf("Max desired start time: %ld\n", max_desired_start_time);
    } else {
        printf("Max desired start time: %s",ctime(&max_desired_start_time));
    }
    printf("Min desired lifetime(seconds): \t%ld\n",min_desired_lifetime);
    printf("Max desired lifetime(seconds): \t%ld\n",max_desired_lifetime);   
}
 

/* set new time values(all) */
static int set_desired_times(){
    struct tm tmin,tmax;
    char *word;
    char *separate = " \t";    
    int month = 0;   

    /* setting min_desired_start_time */
    printf("Set max/min starttime time on format: hh mm ss day month\n");

    cmd = readline("Set min desired start time (press enter for no time): ");
    if(*cmd != 0) {
        add_history(cmd);
        word       = strtok(cmd,separate);
        tmin.tm_hour = atoi(word);
        word       = strtok(NULL,separate);
        tmin.tm_min  = atoi(word);
        word       = strtok(NULL,separate);
        tmin.tm_sec  = atoi(word);
        word       = strtok(NULL,separate);
        tmin.tm_mday = atoi(word);
        word       = strtok(NULL,separate);
        month      = atoi(word);
        tmin.tm_mon  = month-1;
        tmin.tm_year = start_year;
        min_desired_start_time  = mktime(&tmin);
        printf("New min desired start time: %s",ctime(&min_desired_start_time));
    } else {
        min_desired_start_time  = 0;
        printf("New min desired start time: %ld\n", min_desired_start_time);
    }

    /* setting max_desired_start_time */
    cmd = readline("Set max desired start time (press enter for no time): ");
    if(*cmd != 0) {
        add_history(cmd);
        word       = strtok(cmd,separate);
        tmax.tm_hour = atoi(word);
        word       = strtok(NULL,separate);
        tmax.tm_min  = atoi(word);
        word       = strtok(NULL,separate);
        tmax.tm_sec  = atoi(word);
        word       = strtok(NULL,separate);
        tmax.tm_mday = atoi(word);
        word       = strtok(NULL,separate);
        month      = atoi(word);
        tmax.tm_mon  = month-1;
        tmax.tm_year = start_year;    
        max_desired_start_time  = mktime(&tmax);
        printf("New max desired start time: %s",ctime(&max_desired_start_time));
    } else {
        max_desired_start_time  = 0;
        printf("New max desired start time: %ld\n", max_desired_start_time);
    }
    
    /* setting min_desired_lifetime */   
    cmd = readline("Set min desired lifetime(seconds): ");
    add_history(cmd);
    min_desired_lifetime = atoi(cmd);
       
    /* setting max_desired_lifetime */
    cmd = readline("Set max desired lifetime(seconds): ");
    add_history(cmd);
    max_desired_lifetime = atoi(cmd);
    return 0;
}


/* set the default scope to be used */
static int set_default_scope(void){
    /* MAAS configuration trab */
    ar.start           = ntohl(inet_addr("239.255.0.0"));
    ar.end             = ntohl(inet_addr("239.255.255.255"));
    
    /* AAP configuration adam. */
    /* ar.start           = ntohl(inet_addr("224.0.0.0")); */
    /* ar.end             = ntohl(inet_addr("224.255.255.255")); */
    names.language_tag = NULL;
    names.name         = NULL;
    sc.addresses       = ar;
    sc.name_set        = &names;
    sc.num_names       = 0;
    sc.ttl             = 16;
    return 0;
}

/* Set current scope */
static int set_scope(void){
    cmd = readline("Enter start address (in standard dot notation): ");
    add_history(cmd);    
    ar.start = ntohl(inet_addr(cmd));
    
    cmd = readline("Enter end address (in standard dot notation): ");
    add_history(cmd);
    ar.end = ntohl(inet_addr(cmd));
    
    cmd = readline("Enter TTL: ");
    add_history(cmd);
    sc.ttl = atoi(cmd);
    sc.addresses = ar;    
  
    return 0;
}

/* view the scope */
static int list_scope(void){
    struct in_addr tmp;
    tmp.s_addr = ntohl(sc.addresses.start);   
    printf("Start address: \t%s\n",inet_ntoa(tmp));
    tmp.s_addr = ntohl(sc.addresses.end);   
    printf("End address: \t%s\n",inet_ntoa(tmp));

    /* TODO: Finish this... */
    printf("Language tag: \t%s\n",sc.name_set->language_tag);
    printf("Name: \t\t%s\n",sc.name_set->name);
    printf("No. of names: \t%d\n",sc.num_names);
    printf("TTL: \t\t%d\n",sc.ttl);
    return 0;
}


/*
 * Prints allocated address information(leases).
 */
static int list_leases(void){
    allocation *tmp;
    struct in_addr sa;
    uint8 *lid   = NULL;
    uint32 *addr = NULL;
    time_t life  = 0 ;
    int i        = 0;
    int len      = 0;

    if(dll == NULL)
        return -1;

    if (firstElement(dll)== 0){
        while(i != -1){
            tmp = thisElement(dll);
            
            /* list allocation id */
            printf("Id of allocation record: %d \n",tmp->idofallocation);
            
            /* list maas serverid */
            sa.s_addr = ntohl(tmp->le->server_address);
            printf("Contacted server:\t%s \n",inet_ntoa(sa));

            /* list address family */
            if(tmp->le->address_family == IPV4){
                printf("Address family: \t%s\n","IPV4");
            }else{
                printf("Address family of lease(not supported):\t%d\n",tmp->le->address_family);
            }

            /* list starttime of allocated address(es) */
            printf("Start time of lease:\t%s",ctime(&tmp->le->start_time));
            
            /* list the time the addresses are valid */
            life = tmp->le->lifetime + tmp->le->start_time;
            printf("Valid until:\t\t%s",ctime(&life));

            /* list the lease identifier */
            lid = tmp->le->lease_id;            
            printf("Lease identifier:\t");
            for(len = 0; len < tmp->le->id_len; len++){
                printf("%d ",*lid);
                lid++;
            }

            /* list the allocated address(es) */
            printf("\n");
            addr = tmp->le->addresses;
            printf("Addresses allocated:(%d)\n",tmp->le->num_addr);
            for(len = 0; len < tmp->le->num_addr; len++){
                sa.s_addr = ntohl(*addr);   
                printf(" \t\t\t%s\n",inet_ntoa(sa));
                addr++; 
            }
            printf("\n");
            i = nextElement(dll);
        }
    }else{
        printf("No addresses allocated yet.\n");
    }
    return 0;
}


/*
 * Sets active allocation-session(lease).
 */
static int set_lease(void){
    allocation *tmp = NULL;
    int id    = 0;
    int i     = 0;
    int found = 0;
    
    cmd = readline("Enter id of lease: ");
    add_history(cmd);    
    id = atoi(cmd);   

    if(dll == NULL || (firstElement(dll)== -1)){
        printf("No addresses allocated yet.\n");
        return -1;
    }

    while((i != -1) && !found ){
        tmp = thisElement(dll);
        if(tmp->idofallocation == id){
            found = 1;
            currentLease = tmp;
            break;
        }
        i = nextElement(dll);
    }
    
    if (found == 0){
        printf("No such lease id: %d\n",id);
        printf("Active lease id: %d\n",tmp->idofallocation);
    }else{                 
        printf("Current active lease id: %d\n",tmp->idofallocation);
    }
    return i;
}


/* Fuction for deallocating addresses.
 * The address(es) deallocated will be the ones
 * currently selected.
 */
static int deallocate(){    
    allocation *tmp;
    int ret_val = 0;    
    if (currentLease == NULL){
        printf("No addresses allocated.\n");
        return -1;
    }   
    tmp = thisElement(dll);

    ret_val = deallocate_multicast_addr(tmp->le);

    if (ret_val == -1){
        mc_perror("Could not deallocate address(es)");
    }else{
        printf("Deallocating id: %d\n",tmp->idofallocation);
        /* TODO: free all allocated memory */
        deleteElement(dll);
        /* set currentlease to the following/previous element */
        currentLease = thisElement(dll);
    }
    return 0;    
}



/* List avaiable commands */
static void help(){
    printf("Avaliable commands:\n");
    printf("1:\t\t Execute previous command.\n");
    printf("alloc_server: \t Allocate multicast address(es) from a known server.\n");
    printf("alloc_noserver:  Allocate multicast address(es) from an unknown server.\n");
    printf("alloc_one: \t Allocate only a single multicast address from an unknown server.\n");
    printf("alloc_exact: \t Allocate multicast address(es) with a single, fixed lifetime.\n");
    printf("\t\t 'Min desired lifetime' will be used as lifetime.\n");
    printf("alloc_imm: \t Allocate multicast address(es) for immediate use.\n");

    printf("change_lifetime: Change addresses' lifetime.\n");
    printf("\t\t Current selected lease will be used.\n");
    printf("deallocate: \t Deallocate previously allocated addresses.\n");
    printf("\t\t Current selected lease will be used.\n");

    printf("\n");
    
    printf("set_server: \t Set server to connect to.\n");
    printf("list_server: \t Show server selected to connect to.\n");
    printf("clear_server: \t Clear current server selected to connect to.\n");

    printf("list_addresses:  Shows number of desired addresses wanted.\n");
    printf("set_addresses: \t Set number of desired addresses wanted.\n");
    printf("list_times: \t Shows desired times for min/max start/life time. \n");
    printf("set_times: \t Set desired times for min/max start/life time \n");
    printf("show: \t\t Shows all desired addresses/times wanted.\n");

    printf("list_leases: \t Show all leases with allocated addresses.\n");
    printf("set_lease: \t Set active lease.\n");
    printf("list_scope: \t View desired scope to allocate multicast address(es) from.\n");
    printf("set_scope: \t Set desired scope to allocate multicast address(es) from.\n");

    printf("\n");
    
    printf("discover: \t Send a discover message.\n");

    printf("help: \t\t Shows this help.\n");
    printf("exit: \t\t Exit to system.\n");
}
